package org.dynmap.factions;

import com.massivecraft.factions.Factions;
import com.massivecraft.factions.entity.BoardColl;
import com.massivecraft.factions.entity.Faction;
import com.massivecraft.factions.entity.FactionColl;
import com.massivecraft.factions.entity.MFlag;
import com.massivecraft.factions.entity.MFlagColl;
import com.massivecraft.factions.entity.MPlayer;
import com.massivecraft.factions.entity.Warp;
import com.massivecraft.massivecore.ps.PS;
import com.massivecraft.massivecore.store.EntityInternalMap;
import org.bukkit.ChatColor;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.PluginManager;
import org.bukkit.plugin.java.JavaPlugin;
import org.dynmap.DynmapAPI;
import org.dynmap.markers.AreaMarker;
import org.dynmap.markers.Marker;
import org.dynmap.markers.MarkerAPI;
import org.dynmap.markers.MarkerIcon;
import org.dynmap.markers.MarkerSet;
import org.dynmap.markers.PlayerSet;

import java.io.IOException;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

public class DynmapFactionsPlugin extends JavaPlugin {

    private static Logger log;
    private static final String DEF_INFOWINDOW = "<div class=\"infowindow\"><span style=\"font-size:120%;\">%regionname%</span><br />Flags<br /><span style=\"font-weight:bold;\">%flags%</span></div>";
    Plugin dynmap;
    DynmapAPI api;
    MarkerAPI markerapi;
    Plugin factions;
    Factions factapi;
    boolean playersets;

    int blocksize;

    MarkerSet set;
    long updperiod;
    boolean use3d;
    String infowindow;
    AreaStyle defstyle;
    Map<String, AreaStyle> cusstyle;
    Set<String> visible;
    Set<String> hidden;
    boolean stop;
    static DynmapFactionsPlugin instance;

    @Override
    public void onLoad() {
        log = this.getLogger();
    }

    public static void info(String msg) {
        log.log(Level.INFO, msg);
    }

    public static void severe(String msg) {
        log.log(Level.SEVERE, msg);
    }

    void requestUpdatePlayerSet(String factid) {
        if (playersets) {
            getServer().getScheduler().scheduleSyncDelayedTask(this, new PlayerSetUpdate(this, factid));
        }
    }

    FactionsUpdate pending = null;

    void updatePlayerSet(String factid) {
        /* If Wilderness or other unassociated factions (guid-style ID), skip */
        if (factid.indexOf('-') >= 0) {
            return;
        }
        Set<String> plids = new HashSet<>();
        FactionColl fc = FactionColl.get();

        Faction f = fc.getByName(factid);    /* Get faction */
        if (f != null) {
            List<MPlayer> ps = f.getMPlayers();
            for (MPlayer fp : ps) {
                plids.add(fp.getId());
            }
            factid = f.getId();
        }
        String setid = "factions." + factid;
        PlayerSet set = markerapi.getPlayerSet(setid);  /* See if set exists */
        if ((set == null) && (f != null)) {
            set = markerapi.createPlayerSet(setid, true, plids, false);
            info("Added player visibility set '" + setid + "' for faction " + factid);
        }
        
        if (f != null) {
            set.setPlayers(plids);
        } else if (set != null) {
            set.deleteSet();
        }
    }

    private Map<String, AreaMarker> resareas = new HashMap<>();
    private Map<String, Marker> resmark = new HashMap<>();

    private String formatInfoWindow(Faction fact) {
        String v = "<div class=\"regioninfo\">" + infowindow + "</div>";
        v = v.replace("%regionname%", ChatColor.stripColor(fact.getName()));
        v = v.replace("%description%", ChatColor.stripColor(fact.getDescription()));
        MPlayer adm = fact.getLeader();
        v = v.replace("%playerowners%", (adm != null) ? adm.getName() : "");
        StringBuilder res = new StringBuilder();
        for (MPlayer r : fact.getMPlayers()) {
            if (!res.isEmpty()) {
                res.append(", ");
            }
            res.append(r.getName());
        }
        v = v.replace("%playermembers%", res.toString());

        v = v.replace("%nation%", ChatColor.stripColor(fact.getName()));
        /* Build flags */
        StringBuilder flgs = new StringBuilder();
        for (MFlag ff : MFlagColl.get().getAll()) {
            flgs.append("<br/>").append(ff.getName()).append(": ").append(fact.getFlag(ff));
            v = v.replace("%flag." + ff.getName() + "%", fact.getFlag(ff) ? "true" : "false");
        }
        v = v.replace("%flags%", flgs.toString());
        return v;
    }

    private boolean isVisible(String id, String worldname) {
        if ((visible != null) && (!visible.isEmpty())) {
            if ((!visible.contains(id)) && (!visible.contains("world:" + worldname))) {
                return false;
            }
        }
        if ((hidden != null) && (!hidden.isEmpty())) {
            return !hidden.contains(id) && !hidden.contains("world:" + worldname);
        }
        return true;
    }

    private void addStyle(String resid, AreaMarker m) {
        AreaStyle as = cusstyle.get(resid);
        if (as == null) {
            as = defstyle;
        }
        int sc = 0xFF0000;
        int fc = 0xFF0000;
        try {
            sc = Integer.parseInt(as.strokecolor.substring(1), 16);
            fc = Integer.parseInt(as.fillcolor.substring(1), 16);
        } catch (NumberFormatException ignored) {
        }
        m.setLineStyle(as.strokeweight, as.strokeopacity, sc);
        m.setFillStyle(as.fillopacity, fc);
        m.setBoostFlag(as.boost);
    }

    private MarkerIcon getMarkerIcon(String factname, Faction fact) {
        AreaStyle as = cusstyle.get(factname);
        if (as == null) {
            as = defstyle;
        }
        return as.homeicon;
    }

    /**
     * Find all contiguous blocks, set in target and clear in source
     */
    private int floodFillTarget(TileFlags src, TileFlags dest, int x, int y) {
        int cnt = 0;
        ArrayDeque<int[]> stack = new ArrayDeque<>();
        stack.push(new int[]{x, y});

        while (!stack.isEmpty()) {
            int[] nxt = stack.pop();
            x = nxt[0];
            y = nxt[1];
            if (src.getFlag(x, y)) { /* Set in src */
                src.setFlag(x, y, false);   /* Clear source */
                dest.setFlag(x, y, true);   /* Set in destination */
                cnt++;
                if (src.getFlag(x + 1, y)) {
                    stack.push(new int[]{x + 1, y});
                }
                if (src.getFlag(x - 1, y)) {
                    stack.push(new int[]{x - 1, y});
                }
                if (src.getFlag(x, y + 1)) {
                    stack.push(new int[]{x, y + 1});
                }
                if (src.getFlag(x, y - 1)) {
                    stack.push(new int[]{x, y - 1});
                }
            }
        }
        return cnt;
    }

    /* Handle specific faction on specific world */
    private void handleFactionOnWorld(String factname, Faction fact, String world, LinkedList<FactionBlock> blocks, Map<String, AreaMarker> newmap, Map<String, Marker> newmark) {
        double[] x;
        double[] z;
        int poly_index = 0; /* Index of polygon for given faction */

        /* Build popup */
        String desc = formatInfoWindow(fact);

        /* Handle areas */
        if (isVisible(factname, world)) {
            if (blocks.isEmpty()) {
                return;
            }
            LinkedList<FactionBlock> nodevals = new LinkedList<>();
            TileFlags curblks = new TileFlags();
            /* Loop through blocks: set flags on blockmaps */
            for (FactionBlock b : blocks) {
                curblks.setFlag(b.x(), b.z(), true); /* Set flag for block */
                nodevals.addLast(b);
            }
            /* Loop through until we don't find more areas */
            while (nodevals != null) {
                LinkedList<FactionBlock> ournodes = null;
                LinkedList<FactionBlock> newlist = null;
                TileFlags ourblks = null;
                int minx = Integer.MAX_VALUE;
                int minz = Integer.MAX_VALUE;
                for (FactionBlock node : nodevals) {
                    int nodex = node.x();
                    int nodez = node.z();
                    /* If we need to start shape, and this block is not part of one yet */
                    if ((ourblks == null) && curblks.getFlag(nodex, nodez)) {
                        ourblks = new TileFlags();  /* Create map for shape */
                        ournodes = new LinkedList<>();
                        floodFillTarget(curblks, ourblks, nodex, nodez);   /* Copy shape */
                        ournodes.add(node); /* Add it to our node list */
                        minx = nodex;
                        minz = nodez;
                    }
                    /* If shape found, and we're in it, add to our node list */
                    else if ((ourblks != null) && ourblks.getFlag(nodex, nodez)) {
                        ournodes.add(node);
                        if (nodex < minx) {
                            minx = nodex;
                            minz = nodez;
                        } else if ((nodex == minx) && (nodez < minz)) {
                            minz = nodez;
                        }
                    } else {  /* Else, keep it in the list for the next polygon */
                        if (newlist == null) {
                            newlist = new LinkedList<>();
                        }
                        newlist.add(node);
                    }
                }
                nodevals = newlist; /* Replace list (null if no more to process) */
                if (ourblks != null) {
                    /* Trace outline of blocks - start from minx, minz going to x+ */
                    int cur_x = minx;
                    int cur_z = minz;
                    Direction dir = Direction.XPLUS;
                    ArrayList<int[]> linelist = new ArrayList<>();
                    linelist.add(new int[]{minx, minz}); // Add start point
                    while ((cur_x != minx) || (cur_z != minz) || (dir != Direction.ZMINUS)) {
                        switch (dir) {
                            case XPLUS: /* Segment in X+ direction */
                                if (!ourblks.getFlag(cur_x + 1, cur_z)) { /* Right turn? */
                                    linelist.add(new int[]{cur_x + 1, cur_z}); /* Finish line */
                                    dir = Direction.ZPLUS;  /* Change direction */
                                } else if (!ourblks.getFlag(cur_x + 1, cur_z - 1)) {  /* Straight? */
                                    cur_x++;
                                } else {  /* Left turn */
                                    linelist.add(new int[]{cur_x + 1, cur_z}); /* Finish line */
                                    dir = Direction.ZMINUS;
                                    cur_x++;
                                    cur_z--;
                                }
                                break;
                            case ZPLUS: /* Segment in Z+ direction */
                                if (!ourblks.getFlag(cur_x, cur_z + 1)) { /* Right turn? */
                                    linelist.add(new int[]{cur_x + 1, cur_z + 1}); /* Finish line */
                                    dir = Direction.XMINUS;  /* Change direction */
                                } else if (!ourblks.getFlag(cur_x + 1, cur_z + 1)) {  /* Straight? */
                                    cur_z++;
                                } else {  /* Left turn */
                                    linelist.add(new int[]{cur_x + 1, cur_z + 1}); /* Finish line */
                                    dir = Direction.XPLUS;
                                    cur_x++;
                                    cur_z++;
                                }
                                break;
                            case XMINUS: /* Segment in X- direction */
                                if (!ourblks.getFlag(cur_x - 1, cur_z)) { /* Right turn? */
                                    linelist.add(new int[]{cur_x, cur_z + 1}); /* Finish line */
                                    dir = Direction.ZMINUS;  /* Change direction */
                                } else if (!ourblks.getFlag(cur_x - 1, cur_z + 1)) {  /* Straight? */
                                    cur_x--;
                                } else {  /* Left turn */
                                    linelist.add(new int[]{cur_x, cur_z + 1}); /* Finish line */
                                    dir = Direction.ZPLUS;
                                    cur_x--;
                                    cur_z++;
                                }
                                break;
                            case ZMINUS: /* Segment in Z- direction */
                                if (!ourblks.getFlag(cur_x, cur_z - 1)) { /* Right turn? */
                                    linelist.add(new int[]{cur_x, cur_z}); /* Finish line */
                                    dir = Direction.XPLUS;  /* Change direction */
                                } else if (!ourblks.getFlag(cur_x - 1, cur_z - 1)) {  /* Straight? */
                                    cur_z--;
                                } else {  /* Left turn */
                                    linelist.add(new int[]{cur_x, cur_z}); /* Finish line */
                                    dir = Direction.XMINUS;
                                    cur_x--;
                                    cur_z--;
                                }
                                break;
                        }
                    }
                    /* Build information for specific area */
                    String polyid = factname + "__" + world + "__" + poly_index;
                    int sz = linelist.size();
                    x = new double[sz];
                    z = new double[sz];
                    for (int i = 0; i < sz; i++) {
                        int[] line = linelist.get(i);
                        x[i] = (double) line[0] * (double) blocksize;
                        z[i] = (double) line[1] * (double) blocksize;
                    }
                    /* Find existing one */
                    AreaMarker m = resareas.remove(polyid); /* Existing area? */
                    if (m == null) {
                        m = set.createAreaMarker(polyid, factname, false, world, x, z, false);
                        if (m == null) {
                            info("error adding area marker " + polyid);
                            return;
                        }
                    } else {
                        m.setCornerLocations(x, z); /* Replace corner locations */
                        m.setLabel(factname);   /* Update label */
                    }
                    m.setDescription(desc); /* Set popup */

                    /* Set line and fill properties */
                    addStyle(factname, m);

                    /* Add to map */
                    newmap.put(polyid, m);
                    poly_index++;
                }
            }
        }
    }

    /* Update Factions information */
    void updateFactions() {
        Map<String, AreaMarker> newmap = new HashMap<>(); /* Build new map */
        Map<String, Marker> newmark = new HashMap<>(); /* Build new map */

        /* Parse into faction centric mapping, split by world */
        Map<String, FactionBlocks> blocks_by_faction = new HashMap<>();

        FactionColl fc = FactionColl.get();
        Collection<Faction> facts = fc.getAll();
        for (Faction fact : facts) {
            Set<PS> chunks = BoardColl.get().getChunks(fact);
            String fid = fc.getUniverse() + "_" + fact.getId();
            FactionBlocks factblocks = blocks_by_faction.get(fid); /* Look up faction */
            if (factblocks == null) {    /* Create faction block if first time */
                factblocks = new FactionBlocks();
                blocks_by_faction.put(fid, factblocks);
            }

            for (PS cc : chunks) {
                String world = cc.getWorld();

                /* Get block set for given world */
                LinkedList<FactionBlock> blocks = factblocks.blocks.computeIfAbsent(world, k -> new LinkedList<>());
                FactionBlock fb = new FactionBlock(cc.getChunkX(), cc.getChunkZ());
                blocks.add(fb); /* Add to list */
            }
        }
        /* Loop through factions */
        for (Faction fact : facts) {
            String factname = ChatColor.stripColor(fact.getName());
            String fid = fc.getUniverse() + "_" + fact.getId();
            FactionBlocks factblocks = blocks_by_faction.get(fid); /* Look up faction */
            if (factblocks == null) {
                continue;
            }

            /* Loop through each world that faction has blocks on */
            for (Map.Entry<String, LinkedList<FactionBlock>> worldblocks : factblocks.blocks.entrySet()) {
                handleFactionOnWorld(factname, fact, worldblocks.getKey(), worldblocks.getValue(), newmap, newmark);
            }
            factblocks.blocks.clear();

            /* Now, add marker for home location */
            EntityInternalMap<Warp> warps = fact.getWarps();
            Warp home = null;
            for (String warp : warps.keySet()) {
                if (warp.equalsIgnoreCase("home")) {
                    home = warps.get(warp);
                }
            }
            if (home == null) {
                continue;
            }
            PS homeloc = home.getLocation();
            if (homeloc != null) {
                String markid = fc.getUniverse() + "_" + factname + "__home";
                MarkerIcon ico = getMarkerIcon(factname, fact);
                if (ico != null) {
                    Marker homeMarker = resmark.remove(markid);
                    String lbl = factname + " [home]";
                    if (homeMarker == null) {
                        homeMarker = set.createMarker(markid, lbl, homeloc.getWorld(),
                                homeloc.getLocationX(), homeloc.getLocationY(), homeloc.getLocationZ(), ico, false);
                    } else {
                        homeMarker.setLocation(homeloc.getWorld(), homeloc.getLocationX(), homeloc.getLocationY(), homeloc.getLocationZ());
                        homeMarker.setLabel(lbl);   /* Update label */
                        homeMarker.setMarkerIcon(ico);
                    }
                    if (homeMarker != null) {
                        homeMarker.setDescription(formatInfoWindow(fact)); /* Set popup */
                        newmark.put(markid, homeMarker);
                    }
                }
            }
        }
        blocks_by_faction.clear();

        /* Now, review old map - anything left is gone */
        for (AreaMarker oldm : resareas.values()) {
            oldm.deleteMarker();
        }
        for (Marker oldm : resmark.values()) {
            oldm.deleteMarker();
        }
        /* And replace with new map */
        resareas = newmap;
        resmark = newmark;

    }

    private void updatePlayerSets() {
        if (playersets) {
            FactionColl fc = FactionColl.get();
            for (Faction f : fc.getAll()) {
                if ((f == fc.getNone()) || (f == fc.getWarzone()) || (f == fc.getSafezone())) {
                    continue;
                }
                updatePlayerSet(f.getId());
            }
        }
    }

    public void onEnable() {
        instance = this;
        info("initializing");
        PluginManager pm = getServer().getPluginManager();
        /* Get dynmap */
        dynmap = pm.getPlugin("dynmap");
        if (dynmap == null) {
            severe("Cannot find dynmap!");
            return;
        }
        api = (DynmapAPI) dynmap; /* Get API */
        /* Get Factions */
        Plugin p = pm.getPlugin("Factions");
        if (p == null) {
            severe("Cannot find Factions!");
            return;
        }
        factions = p;

        /* If both enabled, activate */
        if (dynmap.isEnabled() && factions.isEnabled()) {
            activate();
        }

        try {
            MetricsLite ml = new MetricsLite(this);
            ml.start();
        } catch (IOException ignored) {
        }
    }

    private boolean reload = false;

    void activate() {
        markerapi = api.getMarkerAPI();
        if (markerapi == null) {
            severe("Error loading dynmap marker API!");
            return;
        }
        /* Connect to factions API */
        factapi = Factions.get();

        blocksize = 16; /* Fixed at 16 */

        /* Load configuration */
        if (reload) {
            this.reloadConfig();
            if (set != null) {
                set.deleteMarkerSet();
                set = null;
            }
        } else {
            reload = true;
        }
        FileConfiguration cfg = getConfig();
        cfg.options().copyDefaults(true);   /* Load defaults, if needed */
        this.saveConfig();  /* Save updates, if needed */

        /* Now, add marker set for mobs (make it transient) */
        set = markerapi.getMarkerSet("factions.markerset");
        if (set == null) {
            set = markerapi.createMarkerSet("factions.markerset", cfg.getString("layer.name", "Factions"), null, false);
        } else {
            set.setMarkerSetLabel(cfg.getString("layer.name", "Factions"));
        }
        if (set == null) {
            severe("Error creating marker set");
            return;
        }
        /* Make sure these are empty (on reload) */
        resareas.clear();
        resmark.clear();

        int minzoom = cfg.getInt("layer.minzoom", 0);
        if (minzoom > 0) {
            set.setMinZoom(minzoom);
        }
        set.setLayerPriority(cfg.getInt("layer.layerprio", 10));
        set.setHideByDefault(cfg.getBoolean("layer.hidebydefault", false));
        use3d = cfg.getBoolean("use3dregions", false);
        infowindow = cfg.getString("infowindow", DEF_INFOWINDOW);

        /* Get style information */
        defstyle = new AreaStyle(this, cfg, "regionstyle");
        cusstyle = new HashMap<>();
        ConfigurationSection sect = cfg.getConfigurationSection("custstyle");
        if (sect != null) {
            Set<String> ids = sect.getKeys(false);

            for (String id : ids) {
                cusstyle.put(id, new AreaStyle(this, cfg, "custstyle." + id, defstyle));
            }
        }
        List<String> vis = cfg.getStringList("visibleregions");
        if (vis != null) {
            visible = new HashSet<>(vis);
        }
        List<String> hid = cfg.getStringList("hiddenregions");
        if (hid != null) {
            hidden = new HashSet<>(hid);
        }
        /* Chec if player sets enabled */
        playersets = cfg.getBoolean("visibility-by-faction", false);
        if (playersets) {
            try {
                if (!api.testIfPlayerInfoProtected()) {
                    playersets = false;
                    info("Dynmap does not have player-info-protected enabled - visibility-by-faction will have no effect");
                }
            } catch (NoSuchMethodError x) {
                playersets = false;
                info("Dynmap does not support function needed for 'visibility-by-faction' - need to upgrade to 0.60 or later");
            }
        }
        updatePlayerSets();

        /* Set up update job - based on periond */
        int per = cfg.getInt("update.period", 300);
        if (per < 15) {
            per = 15;
        }
        updperiod = (per * 20L);
        stop = false;

        getServer().getScheduler().scheduleSyncDelayedTask(this, new FactionsUpdate(this), 40);   /* First time is 2 seconds */
        getServer().getPluginManager().registerEvents(new OurServerListener(this), this);

        info("version " + this.getDescription().getVersion() + " is activated");
    }

    public void onDisable() {
        if (set != null) {
            set.deleteMarkerSet();
            set = null;
        }
        resareas.clear();
        stop = true;
    }

}
